@@ -3,7 +3,7 @@
 ## aggregators.py
 ##
 ## Author:   Toke Høiland-Jørgensen (toke@toke.dk)
-## Date:     16 oktober 2012
+## Date:     16 October 2012
 ## Copyright (c) 2012, Toke Høiland-Jørgensen
 ##
 ## This program is free software: you can redistribute it and/or modify
@@ -26,19 +26,32 @@ from . import runners, transformers
 
 from .util import classname
 
-from .settings import settings
 import collections
 
+def new(settings):
+    cname = classname(settings.AGGREGATOR, "Aggregator")
+    if not cname in globals():
+        raise RuntimeError("Aggregator not found: '%s'" % settings.AGGREGATOR)
+    try:
+        agg = globals()[cname](settings)
+        for s in list(settings.DATA_SETS.items()):
+            agg.add_instance(*s)
+        return agg
+    except Exception as e:
+        raise RuntimeError("Error loading %s: %s." % (cname, e))
+
+
 class Aggregator(object):
     """Basic aggregator. Runs all jobs and returns their result."""
 
-    def __init__(self):
+    def __init__(self, settings):
         self.instances = {}
         self.threads = {}
-        if settings.LOG_FILE is None:
+        self.settings = settings
+        if self.settings.LOG_FILE is None:
             self.logfile = None
         else:
-            self.logfile = open(settings.LOG_FILE, "a")
+            self.logfile = open(self.settings.LOG_FILE, "a")
 
         self.postprocessors = []
 
@@ -49,7 +62,7 @@ class Aggregator(object):
             instance['delay'] = 0
 
 
-        instance['runner'] = getattr(runners, classname(instance['runner'], 'Runner'))
+        instance['runner'] = runners.get(instance['runner'])
 
         if 'data_transform' in config:
             instance['transformers'] = []
@@ -61,7 +74,7 @@ class Aggregator(object):
         duplicates = config.get('duplicates', None)
         if duplicates is not None:
             for i in range(int(duplicates)-1):
-                self.instances["%s - %d" % (name, i+2)] = instance
+                self.instances["%s::%d" % (name, i+2)] = instance
 
     def aggregate(self):
         raise NotImplementedError()
@@ -74,9 +87,10 @@ class Aggregator(object):
             self.logfile.write("Start run at %s\n" % datetime.now())
 
         result = {}
+        metadata = {}
         try:
             for n,i in list(self.instances.items()):
-                self.threads[n] = i['runner'](n, **i)
+                self.threads[n] = i['runner'](n, self.settings, **i)
                 self.threads[n].start()
             for n,t in list(self.threads.items()):
                 while t.isAlive():
@@ -89,11 +103,20 @@ class Aggregator(object):
                     # post-processor (Avg etc), and should be run as such (by the
                     # postprocess() method)
                     self.postprocessors.append(t.result)
+                elif hasattr(t.result, 'keys'):
+                    for k in t.result.keys():
+                        key = "%s::%s" % (n,k)
+                        result[key] = t.result[k]
+                        if key in self.instances and 'transformers' in self.instances[key]:
+                            for tr in self.instances[key]['transformers']:
+                                result[key] = tr(result[key])
                 else:
                     result[n] = t.result
                     if 'transformers' in self.instances[n]:
                         for tr in self.instances[n]['transformers']:
                             result[n] = tr(result[n])
+                if hasattr(t, 'metadata'):
+                    metadata[n] = t.metadata
         except KeyboardInterrupt:
             self.kill_runners()
             raise
@@ -101,16 +124,11 @@ class Aggregator(object):
         if self.logfile is not None:
             self.logfile.write("Raw aggregated data:\n")
             pprint.pprint(result, self.logfile)
-        return result
+        return result,metadata
 
     def kill_runners(self):
         for t in list(self.threads.values()):
-            t.killed = True
-            if hasattr(t, 'prog'):
-                try:
-                    t.prog.send_signal(signal.SIGINT)
-                except OSError:
-                    pass
+            t.kill()
 
     def postprocess(self, result):
         for p in self.postprocessors:
@@ -132,13 +150,16 @@ class IterationAggregator(Aggregator):
     results. Assumes each job outputs one value."""
 
     def __init__(self, *args, **kwargs):
-        self.iterations = settings.ITERATIONS
         Aggregator.__init__(self, *args, **kwargs)
+        self.iterations = self.settings.ITERATIONS
 
     def aggregate(self, results):
-        results.x_values = list(range(1, self.iterations+1))
         for i in range(self.iterations):
-            results.add_result(i+1, self.collect())
+            data,metadata = self.collect()
+            results.meta('SERIES_META', metadata)
+            if i == 0:
+                results.create_series(data.keys())
+            results.append_datapoint(i, data)
         return results
 
 class TimeseriesAggregator(Aggregator):
@@ -149,14 +170,15 @@ class TimeseriesAggregator(Aggregator):
     values are floating point values."""
 
     def __init__(self, *args, **kwargs):
-        self.step = settings.STEP_SIZE
-        self.max_distance = self.step * 5.0
         Aggregator.__init__(self, *args, **kwargs)
+        self.step = self.settings.STEP_SIZE
+        self.max_distance = self.step * 5.0
 
     def aggregate(self, results):
-        measurements = self.collect()
+        measurements,metadata = self.collect()
         if not measurements:
-            raise RuntimeError("No data to aggregate. Run with -l and check log file to investigate.")
+            raise RuntimeError("No data to aggregate. Run with -L and check log file to investigate.")
+        results.meta('SERIES_META', metadata)
         results.create_series(list(measurements.keys()))
 
         # We start steps at the minimum time value, and do as many steps as are
@@ -165,11 +187,13 @@ class TimeseriesAggregator(Aggregator):
         first_times = [i[0][0] for i in list(measurements.values()) if i and i[0]]
         last_times = [i[-1][0] for i in list(measurements.values()) if i and i[-1]]
         if not (first_times and last_times):
-            raise RuntimeError("No data to aggregate. Run with -l and check log file to investigate.")
+            raise RuntimeError("No data to aggregate. Run with -L and check log file to investigate.")
         t_0 = min(first_times)
         t_max = max(last_times)
         steps = int(math.ceil((t_max-t_0)/self.step))
 
+        results.meta('T0', datetime.fromtimestamp(t_0))
+
         time_labels = []
 
         for s in range(steps):
@@ -190,18 +214,14 @@ class TimeseriesAggregator(Aggregator):
                 t_prev = v_prev = None
                 t_next = v_next = None
 
-                # Some measurements (notably UDP pings) give a spurious value
-                # for the last measurement, so cut off the very last data point
-                # from each series. This should hopefully not lose any valuable
-                # data.
-                for i in range(len(r)-1):
+                for i in range(len(r)):
                     if r[i][0] > t:
                         if i > 0:
                             t_prev,v_prev = r[i-1]
                         else:
                             # minimum interpolation distance on first entry to
                             # avoid multiple interpolations to the same value
-                            max_dist = 0.1
+                            max_dist = self.step*0.5
                         t_next,v_next = r[i]
                         break
                 if t_next is None:
@@ -211,7 +231,7 @@ class TimeseriesAggregator(Aggregator):
                     if t_prev is None:
                         # The first/last data point for this measurement is after the
                         # current t. Don't interpolate, just use the value.
-                        if last and results.last_datapoint(n) == v_next:
+                        if last and results.last_datapoint(n) in (v_next,None):
                             # Avoid repeating last interpolation
                             result[n] = None
                         else:
